import logging
import logging.handlers
import os
import threading

import coloredlogs

from bioat.exceptions import BioatInvalidParameterError

__all__ = ["LoggerManager"]

_logger_cache = {}
_logger_lock = threading.Lock()


class LoggerManager:
    LOG_FORMAT = "%(asctime)s.%(msecs)03d - [%(name)s] - %(filename)s[line:%(lineno)4d] - %(levelname)+8s: %(message)s"
    DEFAULT_LEVEL = os.getenv("BIOAT_LOG_LEVEL", "INFO").upper()

    def __init__(
        self,
        log_level: str = DEFAULT_LEVEL,
        mod_name: str = "bioat.logger",
        cls_name: str | None = None,
        func_name: str | None = None,
    ):
        """Initialize the LoggerManager with default log level and module name.

        Args:
            log_level (str, optional): Default logger level. Defaults to "ERROR".
            mod_name (str, optional): Module name. Defaults to "fbt".
            cls_name (str, optional): The name of the class for which the logger is created.
                Defaults to None.
            func_name (str, optional): The name of the function for which the logger is created.
                Defaults to None.
        """
        self.log_level = self._get_log_level(log_level)
        self.mod_name = mod_name
        self.cls_name = cls_name
        self.func_name = func_name
        self.logger = self._get_or_create_logger()
        # self.logger.propagate = False # TODO

    @staticmethod
    def get_logger(name: str, level: str = DEFAULT_LEVEL) -> logging.Logger:
        """直接返回 logger 实例，用于标准 logging 接口."""
        return LoggerManager(mod_name=name, log_level=level).logger

    def _get_log_level(self, log_level: str) -> int:
        if log_level.upper() not in logging._nameToLevel:
            msg = (
                f"Invalid log level: {log_level}. "
                "Choose from CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET."
            )
            raise BioatInvalidParameterError(
                msg,
            )
        return logging._nameToLevel[log_level.upper()]

    def _get_logger_name(self) -> str:
        parts = [self.mod_name]
        if self.cls_name:
            parts.append(self.cls_name)
        if self.func_name:
            parts.append(self.func_name)
        return ".".join(parts)

    def _get_or_create_logger(self) -> logging.Logger:
        name = self._get_logger_name()
        with _logger_lock:
            if name in _logger_cache:
                return _logger_cache[name]

            logger = logging.getLogger(name)
            logger.setLevel(self.log_level)
            logger.propagate = False

            if not any(isinstance(h, logging.StreamHandler) for h in logger.handlers):
                coloredlogs.install(
                    fmt=self.LOG_FORMAT,
                    level=self.log_level,
                    logger=logger,
                    field_styles={
                        "asctime": {"color": "yellow", "bold": True},
                        "name": {"color": "blue", "bold": True},
                        "filename": {"color": "cyan"},
                        "lineno": {"color": "green", "bold": True},
                        "levelname": {"color": "magenta", "bold": True},
                    },
                    level_styles={
                        "critical": {"color": "red", "bold": True},
                        "error": {"color": "red"},
                        "warning": {"color": "yellow"},
                        "info": {"color": "green"},
                        "debug": {"color": "blue"},
                    },
                )
            _logger_cache[name] = logger
            return logger

    def set_names(self, cls_name: str | None = None, func_name: str | None = None):
        """重新设置类名和函数名，会重新绑定 logger."""
        self.cls_name = cls_name
        self.func_name = func_name
        # 会根据新的 name 自动获取缓存或新建 logger（线程安全）
        # 可能只是应用了同名 logger 的配置,并不会实例化新的 logger而是指向同名旧 logger
        # 也可能 logger 会被新的 logger 实例替换
        self.logger = self._get_or_create_logger()

    def set_level(self, log_level: str):
        """更新日志等级."""
        self.log_level = self._get_log_level(log_level)
        self.logger.setLevel(self.log_level)

    def mute(self):
        """将日志等级设置为 NOTSET."""
        self.set_level(log_level="NOTSET")  # NOTSET is higher than CRITICAL

    def set_file(
        self,
        file: str,
        mode: str = "a",
        max_bytes: int = 10 * 1024 * 1024,
        backup_count: int = 5,
    ):
        """设置普通文件日志."""
        if mode not in {"a", "w"}:
            msg = "Mode must be 'a' or 'w'"
            raise BioatInvalidParameterError(msg)
        os.makedirs(os.path.dirname(file), exist_ok=True)

        self._remove_handler_type(logging.FileHandler)

        if mode == "w" or max_bytes <= 0:
            handler = logging.FileHandler(file, mode=mode)
        else:
            handler = logging.handlers.RotatingFileHandler(
                file,
                mode=mode,
                maxBytes=max_bytes,
                backupCount=backup_count,
            )

        handler.setFormatter(logging.Formatter(self.LOG_FORMAT))
        self.logger.addHandler(handler)

    def _remove_handler_type(self, handler_type):
        """内部方法：移除已有的特定 handler 类型."""
        self.logger.handlers = [
            h for h in self.logger.handlers if not isinstance(h, handler_type)
        ]

    def add_stream_handler(self):
        """添加 console handler（不会重复添加）."""
        if not any(isinstance(h, logging.StreamHandler) for h in self.logger.handlers):
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(logging.Formatter(self.LOG_FORMAT))
            self.logger.addHandler(stream_handler)


if __name__ == "__main__":
    import os
    import tempfile

    print("=== LoggerManager 内部测试开始 ===")

    # 测试 logger 创建与等级
    lm = LoggerManager(mod_name="test.module", log_level="DEBUG")
    assert isinstance(lm.logger, logging.Logger)
    assert lm.logger.level == logging.DEBUG
    print("✅ logger 创建与等级 OK")

    # 测试 set_names
    lm.set_names(cls_name="TestClass", func_name="test_func")
    assert "TestClass" in lm.logger.name
    assert "test_func" in lm.logger.name
    print("✅ set_names OK:", lm.logger.name)

    # 测试 set_level
    lm.set_level("INFO")
    assert lm.logger.level == logging.INFO
    lm.logger.info("✅This is an info message.")
    lm.logger.debug("❌This debug message should not appear.")
    lm.logger.warning("✅This is a warning message.")
    lm.logger.error("✅This is an error message.")
    lm.logger.critical("✅This is a critical message.")
    lm.set_level("DEBUG")
    assert lm.logger.level == logging.DEBUG
    lm.logger.debug("✅This debug message should now appear.")
    lm.logger.info("✅This is another info message after level change.")
    print("✅ set_level OK")

    # 测试非法等级
    try:
        lm.set_level("FAKE_LEVEL")
    except BioatInvalidParameterError as e:
        print("✅ 错误等级捕获 OK:", type(e), e)
    else:
        msg = "❌ 错误等级未正确捕获"
        raise AssertionError(msg)

    # 测试文件日志
    with tempfile.TemporaryDirectory() as tmpdir:
        file_path = os.path.join(tmpdir, "test.log")
        lm.set_file(file_path, mode="w")
        lm.logger.warning("test file log entry")
        with open(file_path) as f:
            content = f.read()
            assert "test file log entry" in content
    print("✅ 文件日志写入 OK:", file_path)

    # 测试 add_stream_handler 不重复
    # 仅统计 StreamHandler 类型的 handler 数量
    def count_stream_handlers(logger):
        return sum(1 for h in logger.handlers if isinstance(h, logging.StreamHandler))

    count_before = count_stream_handlers(lm.logger)
    lm.add_stream_handler()
    lm.add_stream_handler()
    count_after = count_stream_handlers(lm.logger)

    # 至多只能添加一个额外的 StreamHandler（如果 coloredlogs 没装）
    assert count_after == count_before, "Unexpected number of StreamHandlers"
    print(
        "✅ StreamHandler 不重复添加 OK:",
        f"StreamHandlers: before={count_before}, after={count_after}",
    )

    # 静态方式测试
    static_logger = LoggerManager.get_logger("static.test", level="WARNING")
    assert isinstance(static_logger, logging.Logger)
    assert "static.test" in static_logger.name
    print("✅ get_logger 静态方法 OK:", static_logger.name)

    lm = LoggerManager(mod_name="bioat.lib.align", log_level="DEBUG")
    lm.logger.info("Hello Logger!")

    # 修改类名函数名
    lm.set_names(cls_name="MyClass", func_name="run")
    lm.logger.debug("With class+func name.")

    # 静态方式快速获取 logger
    logger = LoggerManager.get_logger("bioat.io", "INFO")
    logger.info("Hello from static logger")
    # 初始化 LoggerManager
    lm = LoggerManager(log_level="debug")
    print("🎉 所有 LoggerManager 内部测试通过！")
    print("_logger_cache:", _logger_cache)  # 打印缓存的 logger 名称



"""TODO."""

import math
import os
import shutil
import subprocess
import sys
import tempfile
from io import StringIO
from pathlib import Path
from typing import cast

import numpy as np
import py3Dmol
from Bio import SeqIO
from Bio.PDB.MMCIFParser import MMCIFParser
from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Structure import Structure as BiopythonStructure
from Bio.PDB.Superimposer import Superimposer
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from bioat.exceptions import (
    BioatFileFormatError,
    BioatFileNotFoundError,
    BioatInvalidParameterError,
)
from bioat.lib.libalignment import (
    get_aligned_seq,
    get_best_alignment,
    instantiate_pairwise_aligner,
)
from bioat.lib.libcolor import map_colors_between_two
from bioat.lib.libfastx import (
    AMINO_ACIDS_1CODE,
    AMINO_ACIDS_3CODE,
    AMINO_ACIDS_3CODE_EXTEND,
)
from bioat.logger import LoggerManager

__all__ = [
    "get_cut2ref_aln_info",
    "load_structure",
    "show_ref_cut",
    "structure2string",
]

lm = LoggerManager(mod_name="bioat.lib.libpdb")
lm.set_level("DEBUG")


def _load_seq(seq: str | Path | Seq, label=None):
    if isinstance(seq, str):
        seq = Path(seq)

    if (
        isinstance(seq, Path)
        and seq.is_file()
        and seq.suffix.lower() in [".fasta", ".fa"]
    ):
        seq = next(SeqIO.parse(seq, "fasta"))
        label = seq.id if label is None else label
        seq = seq.seq
    elif isinstance(seq, Seq):
        label = "no_name" if label is None else label
    else:
        msg = f"Invalid sequence format. seq: {seq}, type={type(seq)}, seq.is_file()={seq.is_file()}, seq.suffix.lower()={seq.suffix.lower()}"
        raise BioatInvalidParameterError(msg)
    return seq, label


def load_structure(
    structure: str | Path | BiopythonStructure, label: str | None = None
) -> tuple[BiopythonStructure, str]:
    """Load a PDB structure file or a BiopythonStructure object.

    Args:
        structure (str | Path | BiopythonStructure): Structure file path or BiopythonStructure object.
        label (str | None, optional): Structure label. If None, the label will be set to the file name. Defaults to None.

    Raises:
        BioatInvalidParameterError: If the input structure is not a valid file path or a BiopythonStructure object.

    Returns:
        tuple[BiopythonStructure, str]: A tuple of BiopythonStructure object and label.
    """
    if isinstance(structure, str):
        structure = Path(structure)

    if isinstance(structure, Path) and structure.is_file():
        parser = PDBParser(QUIET=True)
        structure = cast(BiopythonStructure, parser.get_structure(label, structure))
    if isinstance(structure, BiopythonStructure):
        label = cast(str, structure.id if label is None else label)
        return structure, label
    msg = f"Invalid structure format. structure: {structure}"
    raise BioatInvalidParameterError(msg)


def structure2string(structure: BiopythonStructure) -> str:
    """Convert BiopythonStructure object to string format.

    Args:
        structure (BiopythonStructure): Structure object to convert.

    Returns:
        str: PDB context.
    """
    pdb_io = PDBIO()
    pdb_io.set_structure(structure)
    # Use StringIO to save to memory string
    pdb_string = StringIO()
    pdb_io.save(pdb_string)
    # Return string
    return pdb_string.getvalue()


def pdb2fasta(
    input_pdb: str | Path | BiopythonStructure,
    output_fasta: str | Path | None = None,
    func_return: bool = False,
    log_level: str = "DEBUG",
) -> list[Seq] | None:
    """Converts a PDB file to a FASTA file.

    Details:
        1. **Proteins**:
            The protein sequence for each chain will be extracted as "Chain X Protein".
        2. **DNA and RNA**:
            Bases for DNA (A, T, G, C) will be saved as "Chain X DNA", and bases for RNA (A, U, G, C) will be saved as "Chain X RNA".
        3. **Other molecules**:
            Any unrecognized molecules (e.g., ions, modified molecules) will be labeled as [residue] and stored as "Chain X Other molecules".
        4. **Multi-chain complexes**:
            The program supports multi-chain structures in complexes, and the content of each chain will be recorded separately.

    Args:
        input_pdb (str or Path or BiopythonStructure):
            Path to the input PDB/CIF file or Biopython Structure.
        output_fasta (str or Path, optional):
            Output file path. If None, the output file will be named as the
            basename of the input file with a ".fa" extension. Defaults to None.
        func_return: (bool, optional)
            Whether to return a list of SeqRecord objects, useful when used as a function but not for command line. Defaults to False.
        log_level (str, optional):
            Logging level. Defaults to "WARNING".

    Returns:
        List of SeqRecord if func_return is True, otherwise None.
    """
    lm.set_names(func_name="pdb2fasta")
    lm.set_level(log_level)

    structure = None
    structure_id = "structure"

    # === Handle input file or structure object ===
    if isinstance(input_pdb, str | Path):
        input_pdb = Path(input_pdb)

        if not input_pdb.exists():
            msg = f"Input file not found: {input_pdb}"
            raise BioatFileNotFoundError(msg)

        suffix = input_pdb.suffix.lower()
        parser = (
            MMCIFParser(QUIET=True)
            if suffix.endswith(".cif") or suffix.endswith(".cif.gz")
            else PDBParser(QUIET=True)
            if suffix.endswith(".pdb") or suffix.endswith(".pdb.gz")
            else None
        )
        if parser is None:
            raise BioatFileFormatError("Only .pdb and .cif files are allowed.")

        structure_id = input_pdb.stem
        with input_pdb.open("rt") as f:
            structure = parser.get_structure(structure_id, f)

        if output_fasta is None:
            output_fasta = input_pdb.with_suffix(".fa")
            lm.logger.debug(f"Output is not defined, set output to {output_fasta}")
        else:
            output_fasta = Path(output_fasta)
            if output_fasta.suffix.lower() not in (".fa", ".fasta"):
                raise ValueError("Output file must have a '.fa' / '.fasta' suffix")

    elif isinstance(input_pdb, BiopythonStructure):
        structure = input_pdb
        structure_id = getattr(structure, "id", "structure")
    else:
        msg = f"Invalid input type. Expect Path or BiopythonStructure, got {type(input_pdb)}"
        raise BioatInvalidParameterError(msg)

    lm.logger.debug(f"Processing structure: {structure_id}")

    records = []

    if isinstance(structure, BiopythonStructure):
        for model in structure:
            for chain in model:
                lm.logger.debug(f"Processing chain {chain.id}")
                residues = list(chain)
                if not residues:
                    continue

                res_ids = [res.id[1] for res in residues]
                offset = min(res_ids)
                length = max(res_ids) - offset + 1

                protein = ["-"] * length
                dna = ["-"] * length
                rna = ["-"] * length
                other = ["-"] * length

                for res in residues:
                    idx = res.id[1] - offset
                    name = res.resname.strip().upper()

                    if name in AMINO_ACIDS_3CODE:
                        protein[idx] = AMINO_ACIDS_3CODE[name]
                    elif name in AMINO_ACIDS_3CODE_EXTEND:
                        protein[idx] = AMINO_ACIDS_3CODE_EXTEND[name]
                    elif name in ("DA", "DT", "DG", "DC"):
                        dna[idx] = name[1]
                    elif name in ("A", "U", "G", "C"):
                        rna[idx] = name
                    else:
                        other[idx] = f"[{name}]"

                def _append(seq, label, chain):
                    if any(c != "-" for c in seq):
                        records.append(
                            SeqRecord(
                                Seq("".join(seq)),
                                id=f"{structure_id}|Chain_{chain.id}|{label}",
                                description=f"Chain {chain.id} {label}",
                            )
                        )

                _append(protein, "Protein", chain)
                _append(dna, "DNA", chain)
                _append(rna, "RNA", chain)
                _append(other, "UNKNOWN", chain)
    else:
        raise BioatInvalidParameterError("Invalid structure format.")
    if output_fasta:
        SeqIO.write(records, output_fasta, "fasta")
        lm.logger.info(f"FASTA saved to: {output_fasta}")

    return [s.seq for s in records] if func_return else None


def _map_ref_colors(
    ref_map_colors: tuple,
    ref_map_values: dict,
    ref_map_value_random: bool,
) -> tuple:
    """Map reference residue colors based on values.

    This function maps residue-level scalar values (e.g., importance scores) to colors
    along a linear gradient defined by `ref_map_colors`. Each residue is identified by
    a string key in the format "X_123", where X is the 1-letter amino acid code and 123 is the residue index.

    Args:
        ref_map_colors (tuple): A tuple of two color values (e.g., hex or RGB), representing the color gradient endpoints.
                                For example: ("blue", "red") or ("#0000FF", "#FF0000").
        ref_map_values (dict): A dictionary mapping residue keys to float values.
                               Each key must be in the format "<residue>_<index>", such as "A_25".
                               The value represents a scalar quantity (e.g., attention weight) to be color-mapped.
        ref_map_value_random (bool): If True, the actual mapped values will be replaced with random numbers in [0, 1],
                                     ignoring the values in `ref_map_values`. Useful for debugging or visualization testing.

    Returns:
        tuple:
            - colors (List[str or tuple]): List of color values corresponding to each residue, generated by interpolation.
            - res_names (List[str]): List of 1-letter residue codes.
            - res_idxes (List[int]): List of residue indices (integers parsed from the keys).

    Raises:
        BioatInvalidParameterError:
            - If `ref_map_colors` is not a tuple of length 2.
            - If a residue name in `ref_map_values` is not a valid 1-letter amino acid code.

    Example:
        >>> ref_map_colors = ("blue", "red")
        >>> ref_map_values = {"A_25": 0.1, "V_26": 0.9}
        >>> colors, names, idxes = _map_ref_colors(ref_map_colors, ref_map_values, False)
        >>> print(colors)  # List of interpolated colors between blue and red
        >>> print(names)  # ['A', 'V']
        >>> print(idxes)  # [25, 26]
    """
    if len(ref_map_colors) != 2:
        raise BioatInvalidParameterError("ref_map_colors must be a tuple of length 2")

    color1, color2 = ref_map_colors

    values, res_names, res_idxes = [], [], []
    for k, v in ref_map_values.items():
        res_name, res_idx = k.split("_")
        if res_name not in AMINO_ACIDS_1CODE:
            msg = f"Invalid residue name: {res_name}"
            raise BioatInvalidParameterError(msg)
        res_names.append(res_name)
        res_idxes.append(int(res_idx))
        values.append(v)

    # 映射颜色
    rng = np.random.default_rng()
    mapped_values = rng.random(len(values)) if ref_map_value_random else values

    colors = map_colors_between_two(color1, color2, mapped_values)
    return colors, res_names, res_idxes


def show_ref_cut(
    ref_seq: str | Path | Seq,
    ref_pdb: str | Path | BiopythonStructure,
    cut_seq: list[str | Path | Seq] | str | Path | Seq | None = None,
    cut_pdb: list[str | Path | BiopythonStructure]
    | str
    | Path
    | BiopythonStructure
    | None = None,
    cut_labels: list[str] | str | None = None,
    ref_color: str = "red",
    ref_map_colors: tuple[str, str] | None = None,
    ref_map_values: dict | None = None,
    cut_color="lightgray",
    gap_color="purple",
    ref_style="cartoon",
    cut_style="cartoon",
    gap_style="cartoon",
    ref_map_value_random: bool = False,
    output_fig: str | Path | None = None,
    col: int = 4,
    scale: float = 1.0,
    annotate: bool = True,
    text_interval: int = 5,
    log_level="WARNING",
):
    """Visualizes the alignment of sequences and highlights changes in PDB structures using py3Dmol.

    Args:
        ref_seq (str or Seq): Amino acid sequence content for the ref protein.
        ref_pdb (str or BiopythonStructure): Path to the PDB file of the reference structure.
        cut_seq (str, Seq or None, optional): Amino acid sequence content for the cut protein.
        cut_pdb (str, BiopythonStructure or None, optional): Path to the PDB file of the cut structure.
        cut_labels (list[str] or str or None, optional): Label for the cut proteins. If None, the label will be set to "cut".
        ref_color (str, optional): Color for reference residues.
        ref_map_colors (tuple[str, str] or None, optional): ref_map_colors will be used as color bar from ref_map_colors[0] to ref_map_colors[1]. If None, do not apply color mapping. Defaults to None.
        ref_map_values (dict or None, optional): A dictionary of values for the ref color map, it will be normalized to the range of [0 - 1]. If None, all residues will be colored with the same color. e.g. ref_map_values = {'V_0': 0.4177215189873418, 'S_1': 0.8185654008438819, 'K_2': 0.9915611814345991, 'G_3': 0.42616033755274263, ...}
        cut_color (str, optional): Color for cut residues.
        gap_color (str, optional): Color for gaps or removed residues.
        ref_style (str, optional): "stick", "sphere", "cartoon", or "line"
        cut_style (str, optional): "stick", "sphere", "cartoon", or "line"
        gap_style (str, optional): "stick", "sphere", "cartoon", or "line"
        ref_map_value_random (bool, optional): If True, ref_map_values will be randomly generated. Defaults to False.
        output_fig (str or Path or None, optional): Output figure file path. If None, the figure will not be saved in html format. Defaults to None.
        col (int, optional): Number of columns for the visualization. Defaults to 3.
        scale (float, optional): Scale factor for the visualization. Defaults to 1.0.
        annotate (bool, optional): Whether to annotate the visualization with labels. Defaults to True.
        text_interval (int, optional): The interval between text annotations. Defaults to 5.
        log_level (str, optional): Log level. Defaults to "WARNING".
    """
    lm.set_names(func_name="show_ref_cut")
    lm.set_level(log_level)
    if isinstance(cut_seq, str) and "," in cut_seq:
        cut_seq = cast(list[str | Path | Seq], cut_seq.split(","))
    if isinstance(cut_pdb, str) and "," in cut_pdb:
        cut_pdb = cast(list[str | Path | BiopythonStructure], cut_pdb.split(","))
    ref_seq, ref_label = _load_seq(ref_seq, "ref")
    ref_pdb, ref_label = load_structure(ref_pdb, ref_label)

    aligner = instantiate_pairwise_aligner(
        scoring_match=5,
        penalty_mismatch=-100,
        penalty_gap_open=-1,
        penalty_gap_extension=0,
        penalty_query_left_gap_score=0,
        penalty_query_right_gap_score=0,
        log_level=log_level,
    )

    if cut_seq:  # not None
        cut_seqs = [cut_seq] if not isinstance(cut_seq, list) else cut_seq
        cut_pdbs = [cut_pdb] if not isinstance(cut_pdb, list) else cut_pdb

        ref_seqs_fix = []
        ref_pdbs_fix = []
        ref_labels = []
        cut_seqs_fix = []
        cut_pdbs_fix = []
        cut_labels_fix = []
        gap_indices = []
        rmsds = []
        alignment_dicts = []

        for idx, _cut_seq in enumerate(cut_seqs):
            ref_seqs_fix.append(ref_seq)
            ref_labels.append(ref_label)
            # 执行Sequence Alignment并得到gaps的位置
            _cut_seq, cut_label = _load_seq(_cut_seq, "cut")
            cut_seqs_fix.append(_cut_seq)

            if cut_labels:
                if isinstance(cut_labels, list):
                    cut_labels_fix.append(cut_labels[idx])
                elif isinstance(cut_labels, str):
                    cut_labels_fix.append(cut_labels)
            else:
                cut_labels_fix.append(cut_label)

            # 获得gaps的位置
            # Perform alignment
            alignment = get_best_alignment(
                seq_a=ref_seq,
                seq_b=_cut_seq,
                aligner=aligner,
                consider_strand=False,
            )
            alignment_dict = get_aligned_seq(alignment, reverse=False, letn_match=False)
            alignment_dicts.append(alignment_dict)
            gap_indices.append(  # 不指定 cut pdb 时必须在这里计算而不是get_cut2ref_aln_info中
                [i for i, n in enumerate(alignment_dict["target_seq"]) if n == "-"],
            )

            cut_pdb = cut_pdbs[idx]

            if cut_pdb:
                # 执行PDB Alignment并得对齐后的PDB文件和RMSD
                res = get_cut2ref_aln_info(
                    ref_pdb,
                    cut_pdb,
                    cal_rmsd=True,
                    cal_tmscore=False,
                    label1=ref_label,
                    label2=cut_label,
                    log_level=log_level,
                )
                ref_pdbs_fix.append(res[ref_label])
                cut_pdbs_fix.append(res[cut_label])
                rmsds.append(res["RMSD"])

                if not isinstance(res[cut_label], BiopythonStructure):
                    msg = "Invalid input format. cut_pdb must be (one or list of) BiopythonStructure"
                    raise BioatInvalidParameterError(msg)
            else:
                ref_pdbs_fix.append(ref_pdb)
                cut_pdbs_fix.append(None)
                rmsds.append(None)
    else:
        lm.logger.info(
            "Cut sequence is not provided, skip alignment and annotation for gaps, and just show ref visualization.",
        )
        ref_seqs_fix = [ref_seq]
        ref_pdbs_fix = [ref_pdb]
        ref_labels = [ref_label]
        cut_seqs_fix = [None]
        cut_pdbs_fix = [None]
        cut_labels_fix = [None]
        gap_indices = [None]
        rmsds = [None]
        alignment_dicts = [None]

    n = len(ref_seqs_fix)
    col = min(n, col)
    row = math.ceil(n / col)

    view = py3Dmol.view(
        width=int(500 * scale * col),
        height=int(500 * scale * row),
        viewergrid=(row, col) if n > 1 else None,
        linked=True,  # 是否同步网格间的分子运动
    )
    lm.logger.debug(f"viewergrid: ({row}, {col})")
    viewers = [(i // col, i % col) for i in range(n)]
    lm.logger.debug(f"viewers = {viewers}")

    for i, viewer in enumerate(viewers):
        lm.logger.debug(f"viewer = {viewer}")
        _add_one_submodel(
            view=view,
            ref_pdb=ref_pdbs_fix[i],
            ref_label=ref_labels[i],
            cut_pdb=cut_pdbs_fix[i],
            cut_label=cut_labels_fix[i],
            ref_color=ref_color,
            ref_map_colors=ref_map_colors,
            ref_map_values=ref_map_values,
            ref_map_value_random=ref_map_value_random,
            cut_color=cut_color,
            gap_color=gap_color,
            gap_indices=gap_indices[i],
            ref_style=ref_style,
            cut_style=cut_style,
            gap_style=gap_style,
            alignment_dict=alignment_dicts[i],
            rmsd=rmsds[i],
            annotate=annotate,
            text_interval=text_interval,
            viewer=viewer,
        )

    if output_fig:
        output_fig = Path(output_fig)
        if output_fig.suffix != ".html":
            lm.logger.warning(
                f"Output file must have a '.html' extension, but got {output_fig}"
            )
            sys.exit(1)
        else:
            with output_fig.open("w") as f:
                f.write(view._make_html())  # noqa: SLF001
            lm.logger.info(f"Visualization saved to: {output_fig}")
    return view.show()


def _add_one_submodel(
    view,
    ref_pdb,
    ref_label,
    cut_pdb,
    cut_label,
    ref_color,
    ref_map_colors,
    ref_map_values,
    ref_map_value_random,
    cut_color,
    gap_color,
    gap_indices,
    ref_style,
    cut_style,
    gap_style,
    alignment_dict,
    rmsd,
    annotate,
    text_interval,
    viewer,
):
    # for annotation
    label_xpos = -70
    label_ypos = 75
    label_zpos = -30

    if ref_map_colors is not None:
        colors, res_names, res_idxes = _map_ref_colors(
            ref_map_colors,
            ref_map_values,
            ref_map_value_random,
        )
    else:
        colors, res_names, res_idxes = None, None, None
    color1, color2 = None, None
    if ref_map_colors is not None:
        color1, color2 = ref_map_colors
    if ref_map_colors and ref_map_values:
        values = []
        res_names = []
        res_idxes = []
        for k, v in ref_map_values.items():
            res_name, res_idx = k.split("_")
            if res_name not in AMINO_ACIDS_1CODE:
                msg = f"Invalid residue name: {res_name}"
                raise BioatInvalidParameterError(msg)
            res_idx = int(res_idx)
            values.append(v)
            res_names.append(res_name)
            res_idxes.append(res_idx)
        if not ref_map_value_random:
            colors = map_colors_between_two(color1, color2, values)
        else:
            rng = np.random.default_rng()
            mapped_values = rng.random(len(values))
            colors = map_colors_between_two(color1, color2, mapped_values)

    view.setViewStyle({"style": "outline", "color": "black", "width": 0.1})

    view.addModel(
        structure2string(ref_pdb),
        "pdb",
        viewer=viewer,
    )
    view.setStyle(
        {"model": 0},
        {
            ref_style: {
                "arrows": True,
                "tubes": True,
                "style": "oval",
                "color": ref_color,
            },
        },
        viewer=viewer,
    )
    label_ypos -= text_interval
    view.addLabel(
        f"{ref_label}: {ref_color}",
        {
            "fontColor": ref_color,
            "fontSize": 14,
            "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
        },
        viewer=viewer,
    )

    # 标记ref 值的颜色
    if (
        ref_map_colors is not None
        and colors is not None
        and res_names is not None
        and res_idxes is not None
    ):
        for i in range(len(colors)):
            idx = res_idxes[i]
            res = res_names[i]
            color = colors[i]
            view.addStyle(
                {"model": 0, "resi": idx + 1, "resn": AMINO_ACIDS_1CODE[res]},
                {ref_style: {"color": color, "radius": 0.5}},
                viewer=viewer,
            )
        if annotate:
            label_ypos -= text_interval
            view.addLabel(
                f"{ref_label} values: {ref_color} (0) to {ref_color} (1)",
                {
                    "fontColor": ref_color,
                    "fontSize": 14,
                    "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
                },
                viewer=viewer,
            )
    # 标记gaps的颜色
    if gap_indices:
        for idx, res in enumerate(alignment_dict["reference_seq"]):
            # 标记 gap位置的颜色
            if idx in gap_indices:
                view.addStyle(
                    {"model": 0, "resi": idx + 1, "resn": AMINO_ACIDS_1CODE[res]},
                    {gap_style: {"color": gap_color, "radius": 0.5}},
                    viewer=viewer,
                )
        if annotate:
            label_ypos -= text_interval
            view.addLabel(
                f"gaps: {gap_color}",
                {
                    "fontColor": gap_color,
                    "fontSize": 14,
                    "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
                },
                viewer=viewer,
            )
    if cut_pdb and rmsd:
        view.addModel(structure2string(cut_pdb), "pdb", viewer=viewer)
        view.setStyle(
            {"model": 1},
            {
                cut_style: {
                    "arrows": True,
                    "tubes": True,
                    "style": "oval",
                    "color": cut_color,
                },
            },
            viewer=viewer,
        )
        if annotate:
            label_ypos -= text_interval
            view.addLabel(
                f"{cut_label}: {cut_color}\tRMSD: {rmsd:.2f}",
                {
                    "fontColor": cut_color,
                    "fontSize": 14,
                    "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
                },
                viewer=viewer,
            )

    # 聚焦显示
    # Finalize the view
    view.zoomTo(viewer=viewer)
    # view.render(viewer=viewer)


# 对齐函数
def get_cut2ref_aln_info(
    ref: str | Path | BiopythonStructure,
    cut: str | Path | BiopythonStructure,
    cal_rmsd=True,
    cal_tmscore=False,
    label1="ref",
    label2="cut",
    usalign_bin: str | Path = "usalign",
    log_level="WARNING",
) -> dict:
    """Align cutted pdb to ref pdb using the CA atoms.

    Aligns a truncated protein structure (cut) to its full-length reference structure (ref)
    using Ca atoms and Biopython's Superimposer.

    This function:
    - Extracts all Ca atoms from `ref` and `cut`
    - Removes atoms from `ref` at the indices listed in `gap_indices`
    - Aligns the remaining atoms from `cut` to the corresponding positions in `ref`
    - Modifies the `cut` structure in-place to match the aligned orientation
    - Returns both structures and the RMSD value of the alignment

    It assumes:
    - One-to-one correspondence between residues after gap removal
    - Structures are predicted by AlphaFold2 / ESMFold (no missing atoms)

    Args:
        ref (str or Bio.PDB.Structure.Structure): Reference structure path or loaded Structure.
        cut (str or Bio.PDB.Structure.Structure): Truncated structure path or loaded Structure.
        cal_rmsd (bool, optional): Whether to calculate RMSD. Default is True.
        cal_tmscore (bool, optional): Whether to calculate TM-score using USalign. Default is False.
        label1 (str, optional): Name for the reference structure. Default is "ref".
        label2 (str, optional): Name for the cut structure. Default is "cut".
        usalign_bin (str or Path, optional): Path to the USalign binary for TM-score calculation. Default is "usalign".
        log_level (str, optional): Logging level. Default is "WARNING".

    Returns:
        dict: {
                "{label1}": aln label1 structure,  # if cal_rmsd is True, unaltered label1 structure
                "{label2}}": fixed label2 structure,  # if cal_rmsd is True, fix label2 coords in-place
                "RMSD": 0.123  # if cal_rmsd is True, the RMSD value between label1 and label2
                f"{label1}_seq": ref_seq,  # if cal_rmsd is True, the sequence of label1 structure
                f"{label2}_seq": cut_seq,  # if cal_rmsd is True, the sequence of label2 structure
                "alignment_dict": alignment_dict,  # if cal_rmsd is True, the alignment dict of label1 and label2
                "gap_indices": gap_indices,  # if cal_rmsd is True, the indices of gaps in label1 structure
                "TM-score:mean": 0.623,  # if cal_tmscore is True, the mean TM-score value
                "TM-score:TM1": 0.456,  # if cal_tmscore is True, use label1 as ref <L_N> in calculation
                "TM-score:TM2": 0.789,  # if cal_tmscore is True, use label2 as ref <L_N> in calculation
                ...
            }
    """
    # 读取 PDB 文件
    ref, cut = (
        load_structure(ref, label1)[0],
        load_structure(cut, label2)[0],
    )
    aln_info = {}

    if cal_rmsd:
        lm.logger.debug(
            f"Calculating RMSD between {label1} and {label2} using CA atoms.",
        )
        ref_seq = cast(Seq, pdb2fasta(ref, func_return=True)[0])  # type: ignore
        cut_seq = cast(Seq, pdb2fasta(cut, func_return=True)[0])  # type: ignore
        lm.logger.debug(f"fetch seq info from structure: {ref.id} -> {ref_seq}")
        lm.logger.debug(f"fetch seq info from structure:  {cut.id} -> {cut_seq}")
        aligner = instantiate_pairwise_aligner(
            scoring_match=5,
            penalty_mismatch=-100,
            penalty_gap_open=-1,
            penalty_gap_extension=0,
            penalty_query_left_gap_score=0,
            penalty_query_right_gap_score=0,
            log_level=log_level,
        )

        # 获得gaps的位置
        # Perform alignment
        alignment = get_best_alignment(
            seq_a=ref_seq,
            seq_b=cut_seq,
            aligner=aligner,
            consider_strand=False,
        )
        alignment_dict = get_aligned_seq(alignment, reverse=False, letn_match=False)
        gap_indices = [
            i for i, n in enumerate(alignment_dict["target_seq"]) if n == "-"
        ]
        # 选择对齐的原子 (e.g., CA)
        atoms1 = [atom for atom in ref.get_atoms() if atom.get_name() == "CA"]
        atoms1 = [atoms1[i] for i in range(len(atoms1)) if i not in gap_indices]
        atoms2 = [atom for atom in cut.get_atoms() if atom.get_name() == "CA"]

        if len(atoms1) != len(atoms2):
            msg = (
                f"The number of CA atoms in {label1} and {label2} are not equal. "
                "Use ref and cut pdbs predicted by AlphaFold2/ESMFold or other tools "
                "but not pdb downloaded from <PDB database> to ensure the CA atoms are aligned."
            )
            raise BioatInvalidParameterError(msg)

        # 使用 Biopython 的 Superimposer 进行对齐
        super_imposer = Superimposer()
        super_imposer.set_atoms(atoms1, atoms2)
        super_imposer.apply(cut.get_atoms())  # 修改 pdb 的坐标
        rmsd = super_imposer.rms
        lm.logger.info(
            f"RMSD (strict, remove gaps in {label1}) between {label1} and {label2}: {rmsd:.3f}",
        )
        aln_info[label1] = ref
        aln_info[label2] = cut  # 坐标已对齐
        aln_info["RMSD"] = (
            rmsd  # it should be close to the result of pymol cmd.align("cut and name CA", "ref and name CA", cutoff=10.0, cycles=0)[0]
        )
        aln_info[f"{label1}_seq"] = str(ref_seq)
        aln_info[f"{label2}_seq"] = str(cut_seq)
        aln_info["alignment_dict"] = alignment_dict
        aln_info["gap_indices"] = gap_indices

    if cal_tmscore:
        lm.logger.debug(
            f"Calculating TM-score between {label1} and {label2} using CA atoms.",
        )
        # 判断 usalign_bin 是否可执行
        if (
            isinstance(usalign_bin, str)
            and "/" not in usalign_bin
            and shutil.which(usalign_bin) is None
        ) or (
            isinstance(usalign_bin, str | Path)
            and "/" in str(usalign_bin)
            and not (
                Path(usalign_bin).expanduser().resolve().is_file()
                and os.access(Path(usalign_bin).expanduser().resolve(), os.X_OK)
            )
        ):
            lm.logger.error(
                f"usalign binary not found in var 'PATH' or not executable: {usalign_bin}, please set the --usalign_bin <path/to/usalign> param or use conda install -c conda-forge usalign to install it!"
            )
            msg = f"usalign binary not found or not executable: {usalign_bin}"
            raise BioatFileNotFoundError(msg)

        def save_structure_to_tempfile(structure):
            io = PDBIO()
            io.set_structure(structure)
            with tempfile.NamedTemporaryFile(delete=False, suffix=".pdb") as tmp:
                io.save(tmp.name)
                return tmp.name

        ref_path = save_structure_to_tempfile(ref)
        cut_path = save_structure_to_tempfile(cut)
        # 使用 usalign 计算 TM-score
        cmd = [str(usalign_bin), ref_path, cut_path, "-outfmt", "2", "-mol", "prot"]
        result = subprocess.run(cmd, capture_output=True, text=True, check=False)  # noqa: S603

        if result.returncode != 0:
            msg = f"USalign 执行失败: \n{result.stderr}"
            raise RuntimeError(msg)

        lines = result.stdout.strip().splitlines()
        if len(lines) < 2:
            msg = "USalign 输出异常, 结果缺失。"
            raise ValueError(msg)

        fields = lines[1].split()
        tm1 = float(fields[2])
        tm2 = float(fields[3])
        aln_info["TM-score:mean"] = (tm1 + tm2) / 2.0
        aln_info["TM-score:TM1"] = tm1
        aln_info["TM-score:TM2"] = tm2
    return aln_info


if __name__ == "__main__":
    ref_seq = "data/pdb/EGFP_4EUL_ref.fa"
    cut_seq = "data/pdb/EGFP_4EUL_cut1.fa"
    ref_pdb = "data/pdb/EGFP_4EUL_ref.pdb"
    cut_pdb = "data/pdb/EGFP_4EUL_cut1.pdb"
    ref_map_values = {
        "M_0": 0.0,
        "V_1": 0.890295358649789,
        "S_2": 0.270042194092827,
        "K_3": 0.24050632911392406,
        "G_4": 0.8523206751054853,
        "E_5": 0.5991561181434599,
        "K_238": 0.5316455696202531,
    }
    show_ref_cut(
        ref_seq=ref_seq,
        ref_pdb=ref_pdb,
        cut_seq=cut_seq,
        cut_pdb=cut_pdb,
        ref_color="red",
        ref_map_colors=("black", "blue"),
        ref_map_values=ref_map_values,
        cut_color="green",
        gap_color="red",
        ref_style="cartoon",
        cut_style="cartoon",
        gap_style="cartoon",
        ref_map_value_random=False,
        output_fig=None,
        log_level="DEBUG",
    )
    aln_info = get_cut2ref_aln_info(
        ref=ref_pdb,
        cut=cut_pdb,
        cal_rmsd=True,
        cal_tmscore=True,
        label1="ref",
        label2="cut",
        usalign_bin="/Users/zhaohuanan/micromamba/envs/ProEvaluator/bin/usalign",
        log_level="DEBUG",
    )
    print(aln_info)



"""_summary_.

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

_description_

example 1:
    bioat list
        <in shell>:
            $ bioat list
        <in python consolo>:
            >>> from bioat.cli import Cli
            >>> bioat = Cli()
            >>> bioat.list()
            >>> print(bioat.list())

example 2:
    _example_
"""

import math

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Rectangle

from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.lib.libcolor")


def plot_colortable(colors, *, ncols=4, sort_colors=True, labels=None):
    cell_width = 212
    cell_height = 22
    swatch_width = 48
    margin = 12

    names = list(colors)

    if not labels:
        labels = names

    n = len(names)
    nrows = math.ceil(n / ncols)

    width = cell_width * 4 + 2 * margin
    height = cell_height * nrows + 2 * margin
    dpi = 72

    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(
        margin / width,
        margin / height,
        (width - margin) / width,
        (height - margin) / height,
    )
    ax.set_xlim(0, cell_width * 4)
    ax.set_ylim(cell_height * (nrows - 0.5), -cell_height / 2.0)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()

    for i, name in enumerate(names):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(
            text_pos_x,
            y,
            labels[i],
            fontsize=14,
            horizontalalignment="left",
            verticalalignment="center",
        )

        ax.add_patch(
            Rectangle(
                xy=(swatch_start_x, y - 9),
                width=swatch_width,
                height=18,
                facecolor=name,
                edgecolor="0.7",
            ),
        )

    return fig


def convert_hex_to_rgb(hex_color: str) -> tuple:
    """Convert HEX color to RGB color.

    :param hex_color: str, like '#FFFFAA'
    :return: tuple, like (255, 255, 170)
    """
    value = hex_color.lstrip("#")
    value_len = len(value)
    return tuple(int(value[index : index + 2], 16) for index in range(0, value_len, 2))


def convert_rgb_to_hex(rgb_color: tuple) -> str:
    """Convert RGB color to HEX color.

    :param rgb_color: tuple, like (255, 255, 170)
    :return: str, like '#FFFFAA'
    """
    return ("#{:02x}{:02x}{:02x}".format(*rgb_color)).upper()


def map_color(value_vec, breaks, color_list):
    """INPUT:
        <value_vec>
            np.array or a list of values.

        <breaks>
            A sorted value list, which can split all num into len(color_list) intervals.
            e.g. [0.01, 0.1, 0.5, 1] make all real num into 5 intervals, (-Inf,0.01], (0.01,0.1], (0.1, 0.5],  (0.5, 1], (1, +Inf]

        <color_list>
            A hex-format color list, which have to match with breaks

    Return:
        <value_color_vec>
            A list map the value_vec with breaks
    """
    value_idx_list = []

    for value in value_vec:
        match_state = False
        for index, break_value in enumerate(breaks):
            if value <= break_value:
                value_idx_list.append(index)
                match_state = True
                break

        if not match_state:
            value_idx_list.append(index + 1)

    return tuple(color_list[col_idx] for col_idx in value_idx_list)


def make_color_list(
    low_color_RGB, high_color_RGB, length_out=20, return_fmt="HEX", log_level="DEBUG"
):
    """INPUT
        <low_color_RGB> <high_color_RGB>
            Format like (210, 179, 150), tuple, list, or np.array.

        <back_format>
            Hex OR RGB

    Return:
        <color_list>
    """
    lm.set_names(func_name="make_color_list")
    lm.set_level(log_level)

    return_fmt = return_fmt.upper()
    supported_fmt = ("HEX", "RGB")

    if return_fmt not in supported_fmt:
        lm.logger.critical(
            f"not supported color format: {return_fmt}\n"
            f"supported_fmt = {supported_fmt}",
        )
    low_color = np.array(low_color_RGB)
    high_color = np.array(high_color_RGB)

    color_list = []
    for index in range(0, length_out + 1):
        rgb_color = [
            abs(i) for i in low_color + (high_color - low_color) // length_out * index
        ]
        if return_fmt == "HEX":
            color_list.append(convert_rgb_to_hex(tuple(rgb_color)))
        else:
            color_list.append(tuple(rgb_color))

    return color_list


def map_colors_between_two(base_color, target_color, values):
    """在两个指定颜色之间根据数值映射生成颜色。.

    Args:
        base_color (str): 起始颜色，格式为 '#RRGGBB'。
        target_color (str): 结束颜色，格式为 '#RRGGBB'。
        values (list or array): 数值数组，范围可为任意值。

    Return:
        list: 对应颜色数组（#RRGGBB 格式）。
    """
    # 创建自定义线性颜色映射
    cmap = mcolors.LinearSegmentedColormap.from_list(
        "custom_cmap",
        [base_color, target_color],
    )

    # 归一化数据到 [0, 1]
    norm = mcolors.Normalize(vmin=min(values), vmax=max(values))

    # 映射数值到颜色
    return [mcolors.rgb2hex(cmap(norm(value))) for value in values]


if __name__ == "__main":
    # %%% test func plot_colortable
    colors = [
        "#64C1E8",
        "#80CED7",
        "#63C7B2",
        "#8E6C88",
        "#CA61C3",
        "#FF958C",
        "#883677",
    ]
    plot_colortable(colors, ncols=1, labels=[1, 2, 3, 4, 5, 6, 7])
    # plt.show()



from dataclasses import dataclass


@dataclass
class Bed:
    chromosome: str
    start: int
    end: int
    name: str
    score: int
    strand: str


@dataclass
class Bam:
    pass


@dataclass(frozen=True)
class Fasta:
    """usage:
    fa = Fasta('this_header', 'AGCTACGTCCCCTGA')
    fa
    print(fa).
    """

    header: str
    sequence: str

    def __eq__(self, other):
        return bool(self.header == other.header and self.sequence == other.sequence)

    def __len__(self):
        return len(self.sequence)

    @property
    def length(self):
        return len(self.sequence)

    def __str__(self, line_width: int = 20):
        ls = []

        for i in range(self.length // line_width + 1):
            ls.append(self.sequence[i * line_width : (i + 1) * line_width])
        s = "\n".join(ls)
        return f">{self.header}\n{s}"

    def __repr__(self, show_all: bool = False, show_length: int = 10):
        if show_all:
            return f"Fasta(header='{self.header}', sequence='{self.sequence}')"
        if self.length > show_length:
            s = (
                self.sequence[: show_length // 2]
                + "..."
                + self.sequence[-show_length // 2 :]
            )
            return f"Fasta(header='{self.header}', sequence='{s}')"
        return f"Fasta(header='{self.header}', sequence='{self.sequence}')"


@dataclass(frozen=True)
class Fastq(Fasta):
    header: str
    sequence: str
    extras: str
    quality: str

    def __eq__(self, other, check_quality: bool = False, check_extras: bool = False):
        # rewrite from class Fasta
        checker = [self.sequence == other.sequence]

        if check_extras:
            checker.append(self.extras == other.extras)
        if check_quality:
            checker.append(self.quality == other.quality)

        return bool(all(checker))

    def __str__(self, line_width: int = 20):
        # rewrite from class Fasta
        return f"@{self.header}\n{self.sequence}\n{self.extras}\n{self.quality}\n"

    def __repr__(self):
        # rewrite from class Fasta
        return (
            f"Fastq(header='{self.header}', sequence='{self.sequence}', "
            f"extras='{self.extras}', quality='{self.quality}')"
        )

    def is_valid(self):
        return self.length == len(self.quality)

    # def get_phread_info(self):
    #     # 判断是Phred33还是Phred64
    #     # https://www.jianshu.com/p/248308513e2e
    #     pass


@dataclass(frozen=True)
class Assembly:
    contigs: list[Fasta]
    length: int
    path: str

    def __init__(self, path: str):
        pass


@dataclass
class VCF:
    pass



"""Some cmdline tools for searching, powered by playwright."""

import importlib
import sys
from datetime import datetime
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd

from bioat.exceptions import BioatInvalidOptionError
from bioat.lib.libpatentseq import query_patent
from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.searchtools")


class SearchTools:
    """Search toolbox."""

    lm.set_names(cls_name="SearchTools")

    def __init__(self):
        pass

    def google_scholar(
        self,
        keyword: str,
        output: str | Path | None = None,
        sort_by: str = "CitePerYear",
        n_results: int = 100,
        plot: bool = False,
        start_year: int | None = None,
        end_year: int = datetime.now().year,
        log_level: str = "WARNING",
        **kwargs,
    ):
        """Search Google Scholar.

        This method creates a DataFrame of publication data from Google Scholar.
        Each result includes title, citations, year, authors, venue, publisher, and link.
        It is useful for finding relevant papers by citation metrics.

        Optionally, it can generate a plot of citation counts versus rank and save the
        table in various formats.

        Args:
            keyword (str): Keyword to search for. For exact matches, wrap in double and single quotes,
                e.g., "'exact keyword'".
            output (str, optional): Output file path. Supported formats: .csv, .tsv, .xls, .xlsx.
                Default is None, which means no output and only print the table in the console.
            sort_by (str): Column to sort the result by, such as "Citations" or "CitePerYear". Default is "CitePerYear".
            n_results (int): Number of search results to retrieve. Default is 100.

            plot (bool): Whether to plot citation count vs. rank. Default is False.
            start_year (int, optional): Optional start year for publication filtering.
            end_year (int): End year for publication filtering. Default is current year.
            log_level (str): Logging level. One of: 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'.

        Returns:
            None. Prints the table and optionally saves or plots it.
        """
        lm.set_names(func_name="google_scholar")
        lm.set_level(log_level)
        lm.logger.info("Run google scholar search...")
        # show params
        lm.logger.info(f"Keyword: {keyword}")
        lm.logger.info(f"Sort by: {sort_by}")
        lm.logger.info(f"Number of results: {n_results}")
        lm.logger.info(f"Start year: {start_year}")
        lm.logger.info(f"End year: {end_year}")
        lm.logger.info(f"Output: {output}")
        lm.logger.info(f"Plot: {plot}")

        try:
            playwright_module = importlib.import_module("playwright.sync_api")
            sync_playwright = playwright_module.sync_playwright
        except ModuleNotFoundError:
            lm.logger.exception(
                "Missing dependency 'playwright'. Please install it by running:\n"
                "    pip install playwright && playwright install"
            )
            sys.exit(1)

        GSCHOLAR_URL = (
            "https://scholar.google.com/scholar?start={}&q={}&hl=en&as_sdt=0,5"
        )
        STARTYEAR_URL = "&as_ylo={}"
        ENDYEAR_URL = "&as_yhi={}"

        def get_year(text: str) -> int:
            for token in text.split():
                if token.strip().isdigit() and 1900 < int(token) <= datetime.now().year:
                    return int(token)
            return 0

        def get_citations(text: str) -> int:
            """Extract citation number from a string like 'Cited by 123'."""
            if "Cited by" in text:
                try:
                    return int(text.split("Cited by")[1].split()[0])
                except Exception:
                    return 0
            return 0

        def get_author(text: str) -> str:
            return text.split("-")[0].strip()

        def get_venue(text: str) -> str:
            parts = text.split("-")
            return parts[-2].strip() if len(parts) >= 2 else ""

        def get_publisher(text: str) -> str:
            parts = text.split("-")
            return parts[-1].strip() if len(parts) >= 2 else ""

        def launch_browser(p, headless=True):
            return p.chromium.launch(headless=headless)

        GSCHOLAR_MAIN_URL = GSCHOLAR_URL
        if start_year:
            GSCHOLAR_MAIN_URL += STARTYEAR_URL.format(start_year)
        if end_year:
            GSCHOLAR_MAIN_URL += ENDYEAR_URL.format(end_year)

        with sync_playwright() as p:
            browser = launch_browser(p, headless=kwargs.get("headless", True))
            context = browser.new_context(
                user_agent=(
                    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                    "AppleWebKit/537.36 (KHTML, like Gecko) "
                    "Chrome/120.0.0.0 Safari/537.36"
                ),
                viewport={"width": 1920, "height": 1080},
                locale="en-US",
                java_script_enabled=True,
            )
            page = context.new_page()

            links, title, citations, year, author, venue, publisher, rank = (
                [],
                [],
                [],
                [],
                [],
                [],
                [],
                [0],
            )

            for n in range(0, n_results, 10):
                url = GSCHOLAR_MAIN_URL.format(str(n), keyword.replace(" ", "+"))
                page.goto(url, timeout=60000)
                page.wait_for_timeout(1000)
                # 检查是否触发验证码：页面空、或包含“unusual traffic”、“not a robot”等文字
                if (
                    not page.locator(".gs_r.gs_or.gs_scl").count()
                    or "unusual traffic" in page.content()
                ):
                    lm.logger.warning(
                        "\nGoogle Scholar 验证触发, 进入人工处理模式...\n"
                    )
                    browser.close()

                    # 重开非 headless 模式
                    browser = launch_browser(p, headless=False)
                    context = browser.new_context(
                        user_agent=(
                            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                            "AppleWebKit/537.36 (KHTML, like Gecko) "
                            "Chrome/120.0.0.0 Safari/537.36"
                        ),
                        viewport={"width": 1920, "height": 1080},
                        locale="en-US",
                        java_script_enabled=True,
                    )
                    page = context.new_page()
                    page.goto(url)
                    input("请在弹出的浏览器中手动完成验证, 完成后按回车继续...")

                items = page.locator(".gs_r.gs_or.gs_scl").all()
                for it in items:
                    try:
                        link = (
                            it.locator("h3 > a").get_attribute("href") or "Unavailable"
                        )
                        ttl = it.locator("h3 > a").inner_text()
                    except Exception:
                        link = "Unavailable"
                        ttl = "No Title"
                    meta = it.locator(".gs_a").inner_text()
                    title.append(ttl)
                    links.append(link)
                    # citations.append(get_citations(it.inner_html()))
                    try:
                        cite_texts = it.locator(".gs_fl").all_inner_texts()
                        citation_line = next(
                            (t for t in cite_texts if "Cited by" in t), ""
                        )
                        citations.append(get_citations(citation_line))
                    except Exception:
                        citations.append(0)
                    year.append(get_year(meta))
                    author.append(get_author(meta))
                    venue.append(get_venue(meta))
                    publisher.append(get_publisher(meta))
                    rank.append(rank[-1] + 1)

            browser.close()

        df = pd.DataFrame(
            zip(author, title, citations, year, publisher, venue, links, strict=False),
            columns=[
                "Author",
                "Title",
                "Citations",
                "Year",
                "Publisher",
                "Venue",
                "Source",
            ],
        )
        df["CitePerYear"] = df["Citations"] / (
            end_year + 1 - df["Year"].replace(0, end_year + 1)
        )
        df["CitePerYear"] = df["CitePerYear"].round().astype(int)

        df = df[
            [
                "Author",
                "Citations",
                "CitePerYear",
                "Year",
                "Venue",
                "Title",
                "Publisher",
                "Source",
            ]
        ]

        try:
            df_sorted = df.sort_values(by=sort_by, ascending=False).reset_index(
                drop=True
            )
        except Exception:
            lm.logger.warning("Sort failed, using default sort by 'Citations'")
            df_sorted = df.sort_values(by="Citations", ascending=False).reset_index(
                drop=True
            )

        if plot:
            try:
                # 保证 citation 来自最终排序结果（而不是原始顺序）
                citation_values = df_sorted["Citations"].tolist()

                plt.figure(figsize=(8, 5))
                plt.plot(
                    range(1, len(citation_values) + 1),
                    citation_values,
                    marker="o",
                    linestyle="None",
                )
                plt.yscale("log")  # 常用: log scale 能更清晰看分布
                plt.grid(True, which="both", linestyle="--", linewidth=0.5)

                plt.ylabel("Number of Citations (log scale)")
                plt.xlabel("Rank")
                plt.title(f"Keyword: {keyword} ({len(citation_values)} results)")
                plt.tight_layout()
                plt.show()
            except Exception as e:
                lm.logger.warning(f"Failed to plot citation distribution: {e}")

        if output:
            lm.logger.info(f"Exporting results to {output}")
            if not output:
                output = f"{keyword.replace(' ', '_')}.xlsx"
            if output.endswith(".csv"):
                df_sorted.to_csv(output, index=False)
            elif output.endswith(".tsv"):
                df_sorted.to_csv(output, sep="\t", index=False)
            elif output.endswith(".xlsx") or output.endswith(".xls"):
                df_sorted.to_excel(output, index=False)
            else:
                raise BioatInvalidOptionError("Unsupported output format: " + output)
        else:
            print(df_sorted)
        if kwargs.get("_for_test", False):
            return df_sorted
        return None

    def query_patent(
        self,
        seq: str,
        query_name: str | None = None,
        username: str | None = None,
        password: str | None = None,
        via_proxy: str | None = None,  # proxy_server = "socks5://127.0.0.1:8235",
        output: str | None = None,
        nobrowser: bool = True,
        retry: int = 3,
        local_browser: str | None = None,
        rm_fail_cookie: bool = False,
        log_level: str = "INFO",
    ):
        """Return a table with a list of patent blast hit from lens.org.

        :param seq: protein sequence, e.g. MCRISQQKK
        :param query_name: queryName in output table
        :param username: ORCID username(usually a mail)
        :param password: ORCID password
        :param via_proxy: like http://127.0.0.1:8234 socks5://127.0.0.1:8235
        :param output: output table.csv/csv.gz
        :param nobrowser: wether or not to open browser for DEBUG
        :param retry: max retry times
        :param local_browser: local firefox browser executable file path
        :param rm_fail_cookie: remove cookies from local if query fail, default is False
        :param log_level: 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
        """
        lm.set_names(func_name="query_patent")
        lm.set_level(log_level)

        lm.logger.info("Run query patent sequence from lens.org...")
        query_patent(
            seq=seq,
            seq_header=query_name,
            username=username,
            password=password,
            proxy_server=via_proxy,
            output=output,
            headless=nobrowser,
            nretry=retry,
            local_browser=local_browser,
            rm_fail_cookie=rm_fail_cookie,
            log_level=log_level,
        )


if __name__ == "__main__":
    tool = SearchTools()
    tool.google_scholar(keyword="prime editing", n_results=20, plot=False)



"""This module contains custom exception classes for the bioat package."""

__module_name__ = "bioat.exceptions"

__all__ = [
    "BioatError",
    "BioatFileFormatError",
    "BioatFileNotCompleteError",
    "BioatFileNotFoundError",
    "BioatInvalidInputError",
    "BioatInvalidOptionError",
    "BioatInvalidParameterError",
    "BioatMissingDependencyError",
    "BioatRuntimeError",
    "BioatValueError",
]


class BioatError(Exception):
    """Base class for all custom exceptions in the bioat package."""

    def __init__(self, *args, **kwargs):
        # 使用列表来收集消息，避免多次字符串拼接
        messages = []
        if args:
            messages.append(" ".join(map(str, args)))  # 将args转换为字符串并连接
        if kwargs:
            messages.append(
                " ".join(f"{k}={v}" for k, v in kwargs.items()),
            )  # 格式化kwargs
        self.msg = "; ".join(messages)

    def __str__(self):
        """Return a string representation of the exception."""
        return f"{self.msg}"  # 返回当前类的名称


class BioatFileFormatError(BioatError):
    """Exception raised for errors in file format."""

    pass


class BioatFileNotCompleteError(BioatError):
    """Exception raised when a file is not complete."""

    pass


class BioatFileNotFoundError(BioatError):
    """Exception raised when a required file is not found."""

    pass


class BioatInvalidInputError(BioatError):
    """Exception raised for invalid input errors."""

    pass


class BioatInvalidOptionError(BioatError):
    """Exception raised for invalid option errors."""

    pass


class BioatInvalidParameterError(BioatError):
    """Exception raised for invalid parameter errors."""

    pass


class BioatMissingDependencyError(BioatError):
    """Exception raised when a required dependency is missing."""

    pass


class BioatRuntimeError(BioatError):
    """Exception raised for runtime errors."""

    pass


class BioatValueError(BioatError):
    """Exception raised for errors related to values."""

    pass



"""TODO."""

from pathlib import Path

from Bio.PDB.Structure import Structure as BiopythonStructure
from Bio.Seq import Seq

from bioat.lib.libpdb import get_cut2ref_aln_info, pdb2fasta, show_ref_cut
from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.foldtools")


class FoldTools:
    """Folding toolbox."""

    def __init__(self):
        pass

    def show_ref_cut(
        self,
        ref_seq: str | Path | Seq,
        ref_pdb: str | Path | BiopythonStructure,
        cut_seq: list[str | Path | Seq] | str | Path | Seq | None = None,
        cut_pdb: list[str | Path | BiopythonStructure]
        | str
        | Path
        | BiopythonStructure
        | None = None,
        cut_labels: list[str] | str | None = None,
        ref_color: str = "red",
        ref_map_colors: tuple[str, str] | None = None,
        ref_map_values: dict | None = None,
        cut_color="lightgray",
        gap_color="purple",
        ref_style="cartoon",
        cut_style="cartoon",
        gap_style="cartoon",
        ref_map_value_random: bool = False,
        output_fig: str | Path | None = None,
        col: int = 4,
        scale: float = 1.0,
        annotate: bool = True,
        text_interval: int = 5,
        log_level="WARNING",
    ):
        """Visualizes the alignment of sequences and highlights changes in PDB structures using py3Dmol.

        Args:
            ref_seq (str or Path or Seq): Amino acid sequence content for the ref protein.
            ref_pdb (str or Path or Bio.PDB.Structure.Structure): Path to the PDB file of the reference structure.
            cut_seq (str, Path or Seq or None, optional): Amino acid sequence content for the cut protein.
            cut_pdb (str, Path or Bio.PDB.Structure.Structure or None, optional): Path to the PDB file of the cut structure.
            cut_labels (list[str] or str or None, optional): Label for the cut proteins. If None, the label will be set to "cut".
            ref_color (str, optional): Color for reference residues.
            ref_map_colors (tuple[str, str] or None, optional): ref_map_colors will be used as color bar from ref_map_colors[0] to ref_map_colors[1]. If None, do not apply color mapping. Defaults to None.
            ref_map_values (dict or None, optional): A dictionary of values for the ref color map, it will be normalized to the range of [0 - 1]. If None, all residues will be colored with the same color. e.g. ref_value_dict = {'V_0': 0.4177215189873418, 'S_1': 0.8185654008438819, 'K_2': 0.9915611814345991, 'G_3': 0.42616033755274263, ...}
            cut_color (str, optional): Color for cut residues.
            gap_color (str, optional): Color for gaps or removed residues.
            ref_style (str, optional): "stick", "sphere", "cartoon", or "line"
            cut_style (str, optional): "stick", "sphere", "cartoon", or "line"
            gap_style (str, optional): "stick", "sphere", "cartoon", or "line"
            ref_map_value_random (bool, optional): If True, ref_value_dict will be randomly generated. Defaults to False.
            output_fig (str or None, optional): Output figure file path. If None, the figure will not be saved in html format. Defaults to None.
            col (int, optional): Number of columns for the visualization. Defaults to 3.
            scale (float, optional): Scale factor for the visualization. Defaults to 1.0.
            annotate (bool, optional): Whether to annotate the visualization with labels. Defaults to True.
            text_interval (int, optional): The interval between text annotations. Defaults to 5.
            log_level (str, optional): Log level. Defaults to "WARNING".
        """
        show_ref_cut(
            ref_seq=ref_seq,
            cut_seq=cut_seq,
            ref_pdb=ref_pdb,
            cut_pdb=cut_pdb,
            cut_labels=cut_labels,
            ref_color=ref_color,
            ref_map_colors=ref_map_colors,
            ref_map_values=ref_map_values,
            cut_color=cut_color,
            gap_color=gap_color,
            ref_style=ref_style,
            cut_style=cut_style,
            gap_style=gap_style,
            ref_map_value_random=ref_map_value_random,
            output_fig=output_fig,
            col=col,
            scale=scale,
            annotate=annotate,
            text_interval=text_interval,
            log_level=log_level,
        )

    def pdb2fasta(
        self,
        input_pdb: str | Path,
        output_fasta: str | Path | None = None,
        log_level="WARNING",
    ):
        """Converts a PDB file to a FASTA file.

        Details:
            1. **Proteins**:
               The protein sequence for each chain will be extracted as "Chain X Protein".
            2. **DNA and RNA**:
               Bases for DNA (A, T, G, C) will be saved as "Chain X DNA", and bases for RNA (A, U, G, C) will be saved as "Chain X RNA".
            3. **Other molecules**:
               Any unrecognized molecules (e.g., ions, modified molecules) will be labeled as [residue] and stored as "Chain X Other molecules".
            4. **Multi-chain complexes**:
               The program supports multi-chain structures in complexes, and the content of each chain will be recorded separately.

        Args:
            input_pdb (str or Path):
                Path to the input PDB/CIF file or Biopython Structure.
            output_fasta (str or Path, optional):
                Output file path. If None, the output file will be named as the
                basename of the input file with a ".fa" extension. Defaults to None.
            func_return: (bool, optional)
                Whether to return a list of SeqRecord objects, useful when used as a function but not for command line. Defaults to False.
            log_level (str, optional):
                Logging level. Defaults to "WARNING".

        Returns:
            List of SeqRecord if func_return is True, otherwise None.
        """
        pdb2fasta(input_pdb=input_pdb, output_fasta=output_fasta, log_level=log_level)

    def get_cut2ref_aln_info(
        self,
        ref: str | BiopythonStructure,
        cut: str | BiopythonStructure,
        cal_rmsd=True,
        cal_tmscore=False,
        label1="ref",
        label2="cut",
        usalign_bin: str = "usalign",
        log_level="WARNING",
    ):
        """Align cutted pdb to ref pdb using the CA atoms.

        Aligns a truncated protein structure (cut) to its full-length reference structure (ref)
        using Ca atoms and Biopython's Superimposer.

        This function:
        - Extracts all Ca atoms from `ref` and `cut`
        - Removes atoms from `ref` at the indices listed in `gap_indices`
        - Aligns the remaining atoms from `cut` to the corresponding positions in `ref`
        - Modifies the `cut` structure in-place to match the aligned orientation
        - Returns both structures and the RMSD value of the alignment

        It assumes:
        - One-to-one correspondence between residues after gap removal
        - Structures are predicted by AlphaFold2 / ESMFold (no missing atoms)

        Args:
            ref (str or Bio.PDB.Structure.Structure): Reference structure path or loaded Structure.
            cut (str or Bio.PDB.Structure.Structure): Truncated structure path or loaded Structure.
            cal_rmsd (bool, optional): Whether to calculate RMSD. Default is True.
            cal_tmscore (bool, optional): Whether to calculate TM-score using USalign. Default is False.
            label1 (str, optional): Name for the reference structure. Default is "ref".
            label2 (str, optional): Name for the cut structure. Default is "cut".
            usalign_bin (str, optional): Path to the USalign binary for TM-score calculation. Default is "usalign".
            log_level (str, optional): Logging level. Default is "WARNING".

        Returns:
            dict: {
                    "{label1}": aln label1 structure,  # if cal_rmsd is True, unaltered label1 structure
                    "{label2}}": fixed label2 structure,  # if cal_rmsd is True, fix label2 coords in-place
                    "RMSD": 0.123  # if cal_rmsd is True, the RMSD value between label1 and label2
                    f"{label1}_seq": ref_seq,  # if cal_rmsd is True, the sequence of label1 structure
                    f"{label2}_seq": cut_seq,  # if cal_rmsd is True, the sequence of label2 structure
                    "alignment_dict": alignment_dict,  # if cal_rmsd is True, the alignment dict of label1 and label2
                    "gap_indices": gap_indices,  # if cal_rmsd is True, the indices of gaps in label1 structure
                    "TM-score:mean": 0.623,  # if cal_tmscore is True, the mean TM-score value
                    "TM-score:TM1": 0.456,  # if cal_tmscore is True, use label1 as ref <L_N> in calculation
                    "TM-score:TM2": 0.789,  # if cal_tmscore is True, use label2 as ref <L_N> in calculation
                    ...
                }
        """
        aln_info = get_cut2ref_aln_info(
            ref=ref,
            cut=cut,
            cal_rmsd=cal_rmsd,
            cal_tmscore=cal_tmscore,
            label1=label1,
            label2=label2,
            usalign_bin=usalign_bin,
            log_level=log_level,
        )
        for key, value in aln_info.items():
            print(f"{key}: {value}")



import pandas as pd
from Bio.Align import PairwiseAligner
from Bio.Seq import Seq

from bioat.lib.libalignment import get_alignment_info
from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.lib.libcrispr")

# TARGET_REGIONS_LIB
#     for target_seq alignment
#
# bioat target_seq region_heatmap test_sorted.mpileup.info.tsv test.pdf --get_built_in_target_region
#
# INFO:root:You can use <key> in built-in <target_regions> to represent your target_region:
#         <key>   <target_region>
#         EMX1    GAGTCCGAGCAGAAGAAGAA^NGG^
#         HEK3    GGCCCAGACTGAGCACGTGA^NGG^
#         HEK4    GGCACTGCGGCTGGAGGTGG^NGG^
#         ...

TARGET_SEQ_LIB = {
    "EMX1": "GAGTCCGAGCAGAAGAAGAA^NGG^",
    "HEK3": "GGCCCAGACTGAGCACGTGA^NGG^",
    "HEK4": "GGCACTGCGGCTGGAGGTGG^NGG^",
    "RNF2": "GTCATCTTAGTCATTACCTG^NGG^",
    "VEGFA": "GACCCCCTCCACCCCGCCTC^NGG^",
    "CDKN2A": "^TTTA^GCCCCAATAATCCCCACATGTCA",
    "DYRK1A": "^TTTA^GAAGCACATCAAGGACATTCTAA",
    "ND4": "TGCTAGTAACCACGTTCTCCTGATCAAATATCACTCTCCTACTTACAGGA",
    "ND5.1": "TAGCATTAGCAGGAATACCTTTCCTCACAGGTTTCTACTCCAAAGA",
    "ND6": "TGACCCCCATGCCTCAGGATACTCCTCAATAGCCATCG",
}


def cmp_align_list(aln_a: dict, aln_b: dict):
    """Compare function for alignment lists.

    This function processes alignment data and performs comparisons based on the input structure.

    Args:
        input_data (list | dict):
            The alignment data to be processed. Can be provided as:
            - A list, e.g.: [17, 3, 0, 73.0, 1, 22, 'AAGAAGAAGACGAGTCTGCA', '||||||||||||||X|||XX', 'AAGAAGAAGACGAGCCTGAG']
            - A dictionary, e.g.: {'match_count': 26, 'mismatch_count': 3, 'gap_count': 4, 'aln_score': 96.0, 'ref_aln_start': 0, 'ref_aln_end': 32, 'alignment': {'reference_seq': 'GGCACTGCGGCTGGAAAAAAAAAAAAAAA--GT','aln_info': '--..|.|||||||||||||||||||||||--||', 'target_seq': '--GGCAGCGGCTGGAAAAAAAAAAAAAAAAGGT'}}

    Returns:
        None: The function does not return a value but performs comparisons.

    Example:
        >>> compare_alignments(
        ...     [
        ...         17,
        ...         3,
        ...         0,
        ...         73.0,
        ...         1,
        ...         22,
        ...         "AAGAAGAAGACGAGTCTGCA",
        ...         "||||||||||||||X|||XX",
        ...         "AAGAAGAAGACGAGCCTGAG",
        ...     ]
        ... )
    """

    # sort reason score -> gap -> mismatch -> match
    def sign_value(x):
        """HELP
        sign function.
        """
        if x > 0:
            return 1
        if x < 0:
            return -1
        return 0

    align_alphabet_score = {"-": 0, ".": 1, "|": 2}
    sort_keys = ["alignment_score", "gap_count", "mismatch_count", "match_count"]
    sort_rev_state_list = [True, False, False, True]

    for idx, key in enumerate(sort_keys):
        if aln_a[key] - aln_b[key] == 0:
            continue
        if sort_rev_state_list[idx]:
            return -1 * sign_value(aln_a[key] - aln_b[key])
        return sign_value(aln_a[key] - aln_b[key])

    aln_info_a = aln_a["alignment"]["aln_info"]
    aln_info_b = aln_b["alignment"]["aln_info"]

    for idx, char_a in enumerate(aln_info_a):
        if idx <= (len(aln_info_b) - 1):
            value_a = align_alphabet_score[char_a]
            value_b = align_alphabet_score[char_a]
            return sign_value(value_a - value_b)
        pass
    return None


def run_target_seq_align(
    ref_seq: Seq,
    target_seq: Seq,
    aligner: PairwiseAligner,
    PAM: dict | None = None,
    log_level="WARNING",
) -> list:
    """Perform global alignment for the target sequence.

    This function performs global alignment between a reference sequence and a target sequence (usually the sgRNA sequence without PAM) using the specified aligner.

    Args:
        ref_seq (Seq):
            A Seq object from BioPython representing the reference sequence
            for the targeted deep sequencing.
        target_seq (Seq):
            A Seq object from BioPython representing the target region sequence
            for the editing window (usually the sgRNA sequence without PAM).
        aligner (object):
            A PairwiseAligner object from BioPython used for performing the alignment.
        PAM (dict, optional):
            A dictionary containing PAM information with the following keys:
            - 'PAM' (str): The PAM sequence (e.g., "AGG").
            - 'position' (int): The insertion site of `target_seq`.
            - 'weight' (float): The priority weight for PAM alignment. The alignment score will be multiplied by this weight.
            - Example: {'PAM': 'AGG', 'position': 20, 'weight': 1.0}.

    Returns:
        list[dict]:
            A list of dictionaries, each containing alignment results. Each dictionary has the following keys:
            - 'match_count' (int): The number of matching bases in the alignment.
            - 'mismatch_count' (int): The number of mismatched bases.
            - 'gap_count' (int): The number of gaps.
            - 'aln_score' (float): The alignment score.
            - 'ref_aln_start' (int): The starting position of the alignment on the reference sequence.
            - 'ref_aln_end' (int): The ending position of the alignment on the reference sequence.
            - 'alignment' (dict): The alignment details containing:'reference_seq' (str): The aligned reference sequence.'aln_info' (str): The alignment information (e.g., matches, mismatches, and gaps).'target_seq' (str): The aligned target sequence.

    Example:
        An example alignment result might look like this:[{'match_count': 26,'mismatch_count': 3,'gap_count': 4,'aln_score': 96.0,'ref_aln_start': 0,'ref_aln_end': 32,'alignment': {'reference_seq': 'GGCACTGCGGCTGGAAAAAAAAAAAAAAA--GT','aln_info': '--..|.|||||||||||||||||||||||--||','target_seq': '--GGCAGCGGCTGGAAAAAAAAAAAAAAAAGGT'}}]

    """
    lm.set_names(func_name="run_target_seq_align")
    lm.set_level(log_level)

    if PAM:
        # considering PAM
        # DNA info https://www.bioinformatics.org/sms/iupac.html
        alignments_pam = aligner.align(ref_seq, Seq(PAM["PAM"]))
        # use pam to find aim sequence on ref and then use target_seq to align them and give it a higher priority
        dt_PAMs = {}

        for _idx, alignment in enumerate(alignments_pam):
            alignment_analysis_result = get_alignment_info(alignment, reverse=False)
            lm.logger.debug(
                f"alignment_analysis_result for PAM:\n{alignment_analysis_result}"
            )
            ref_aln_start = alignment_analysis_result["ref_aln_start"]
            ref_aln_end = alignment_analysis_result["ref_aln_end"]
            aln_score = alignment_analysis_result["aln_score"] * PAM["weight"]
            dt_PAMs[f"{ref_aln_start}_{ref_aln_end}"] = {
                "aln_score": aln_score * PAM["weight"]
            }

        lm.logger.debug(f"find PAMs on reference: {dt_PAMs}")

        # add PAM left or right or any position
        target_seq = (
            target_seq[: PAM["position"]]
            + Seq(PAM["PAM"])
            + target_seq[PAM["position"] :]
        )
        lm.logger.info(
            f"find PAM on target_seq, target_seq is updated to: {target_seq!s}"
        )

    # alignment part
    alignments = aligner.align(ref_seq, target_seq)
    # parse alignment
    alignment_results = []

    for alignment in alignments:
        alignment_analysis_result = get_alignment_info(alignment, reverse=False)
        lm.logger.debug(f"alignment_analysis_result:\n{alignment_analysis_result}")
        alignment_results.append(alignment_analysis_result)

    if len(alignment_results) == 1:
        lm.logger.debug(f"find 1 alignment result:\n{alignment_results}")
        return alignment_results[0]
    lm.logger.debug(
        f"find {len(alignment_results)} alignment results:\n{alignment_results}"
    )
    df = pd.DataFrame(alignment_results)
    # if contain multiple alignment result with the same score, keep the best one;
    # sort reason score -> gap -> mismatch -> match
    df = df.sort_values(
        by=["aln_score", "gap_count", "mismatch_count", "match_count"],
        ascending=[False, True, True, False],
    )
    return df.iloc[0, :].to_dict()



"""bioat.lib.libplot.

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

This module provides functions for plotting.

example 1:
    init_matplotlib
        <in python consolo>:
            >>> import matplotlib.pyplot as plt
            >>> from bioat.lib.libplot import init_matplotlib
            >>> init_matplotlib(log_level="info")
            >>> plt.plot([1, 2, 3], [4, 5, 6])
            >>> plt.show()

example 2:
    plot_colortable
        <in python consolo>:
            >>> from bioat.lib.libplot import plot_colortable
            >>> colors = ["#64C1E8", "#80CED7", "#63C7B2", "#8E6C88", "#CA61C3", "#FF958C", "#883677"]
            >>> plot_colortable(colors, ncols=1, labels=[1, 2, 3, 4, 5, 6, 7])
            >>> plt.show()
"""

import math
import os
import shutil
import sys
from glob import glob

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from matplotlib.patches import Rectangle

from bioat.exceptions import BioatError, BioatRuntimeError
from bioat.lib.libpath import HOME
from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.lib.libplot")

__all__ = ["init_matplotlib", "plot_colortable"]

BIOAT_DEFAULT_FONTS_DATAPATH = os.path.join(os.path.dirname(__file__), "libplot")
BIOAT_DEFAULT_FONTS = [
    "Helvetica-Bold.ttf",
    "Helvetica-BoldOblique.ttf",
    "Helvetica-Light.ttf",
    "Helvetica-Oblique.ttf",
    "Helvetica.ttf",
]


def _copy_fonts(refresh=False, log_level="warning"):
    """Copy fonts from bioat package to matplotlib package.

    :param log_level: log level for logger
    :type log_level: string
    :raises FileNotFoundError: if not found which site-packages to copy fonts to
    """
    lm.set_names(func_name="_copy_fonts")
    lm.set_level(log_level)

    if refresh:
        files_to_remove = glob(os.path.join(HOME, ".cache", "matplotlib", "fontlist*"))
        for file in files_to_remove:
            os.remove(file)
    try:
        to_path = None
        for i in sys.path:
            if i.endswith("site-packages"):
                to_path = i
                break
        if not to_path:
            lm.logger.warning(BioatRuntimeError("site-packages not found in sys.path"))
            return
        to_path = next(i for i in sys.path if i.endswith("site-packages"))
        to_path = os.path.join(
            to_path,
            "matplotlib",
            "mpl-data",
            "fonts",
            "ttf",
        )
        from_path = BIOAT_DEFAULT_FONTS_DATAPATH
        lm.logger.debug(f"Copying fonts from {from_path} to {to_path}")
        for font in BIOAT_DEFAULT_FONTS:
            shutil.copyfile(
                os.path.join(from_path, font),
                os.path.join(to_path, font),
            )
        lm.logger.debug("Fonts copied successfully")
    except Exception as e:
        lm.logger.exception(BioatError(f"Failed to copy fonts: {e}"))


def init_matplotlib(
    style="seaborn-v0_8-white",
    font="Helvetica",
    refresh=False,
    sns_context="paper",
    sns_style="whitegrid",
    sns_palette="deep",
    sns_font_scale=1.2,
    log_level="INFO",
    **kwargs,
):
    """Easily set matplotlib style.

    :param style: matplotlib style, defaults to 'seaborn-v0_8-white'
    :type style: str, optional
    :param font: use what font in matplotlib, defaults to 'Helvetica'
    :type font: str, optional
    :param refresh: wether to remove matplotlib font cache and reset it, defaults to False
    :type refresh: bool, optional
    :param log_level: log level, defaults to 'INFO'
    :type log_level: str, optional
    :param set_backend_pdf: whether to use core fonts for the PDF backend, defaults to True
    :type set_backend_pdf: bool, optional
    :param set_backend_ps: whether to use core fonts for the PS backend, defaults to True
    :type set_backend_ps: bool, optional
    :param set_backend_svg: whether to use 'none' to replace 'path' (use font but not plot path for characters)for the SVG backend, defaults to 'none'
    :type set_backend_svg: str, optional
    :raises BioatError: if failed to copy fonts
    :raises BioatRuntimeError: if site-packages not found in sys.path
    """
    lm.set_names(func_name="init_matplotlib")
    lm.set_level(log_level)

    lm.logger.info("Initializing matplotlib")
    _copy_fonts(refresh=refresh, log_level=log_level)
    lm.logger.info(
        f"set: plt.style.use('{style}')\n"
        "# set matplotlib style theme\n"
        "# ref: https://matplotlib.org/stable/api/style_api.html",
    )
    plt.style.use(style)

    lm.logger.info(f"set: plt.rcParams['font.family'] = '{font}'")
    plt.rcParams["font.family"] = font
    lm.logger.info(f"set: plt.rcParams['font.sans-serif'] = ['{font}']")
    plt.rcParams["font.sans-serif"] = [font]

    # set backends
    set_backend_pdf = kwargs.get("set_backend_pdf", True)
    set_backend_ps = kwargs.get("set_backend_ps", True)
    set_backend_svg = kwargs.get("set_backend_svg", "none")

    if set_backend_pdf:
        lm.logger.info(
            f"set: plt.rcParams['pdf.use14corefonts'] = {set_backend_pdf} \n"
            "# whether to use core fonts for the PDF backend\n"
            "# ref: https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.rcParams)",
        )
        plt.rcParams["pdf.use14corefonts"] = set_backend_pdf
    if set_backend_ps:
        lm.logger.info(
            f"set: plt.rcParams['ps.useafm'] = {set_backend_ps}\n"
            "# whether to use core fonts for the PS backend",
        )
        plt.rcParams["ps.useafm"] = set_backend_ps
    if set_backend_svg:
        lm.logger.info(
            f"set: plt.rcParams['svg.fonttype'] = '{set_backend_svg}'\n"
            "# whether to use 'none' to replace 'path' (use font but not plot path for characters)for the SVG backend",
        )
        plt.rcParams["svg.fonttype"] = set_backend_svg
    lm.logger.info("matplotlib initialized successfully")

    sns_context = kwargs.get("sns_context", "paper")
    sns_style = kwargs.get("sns_style", "whitegrid")
    sns_palette = kwargs.get("sns_palette", "deep")
    sns_font_scale = kwargs.get("sns_font_scale", 1.2)
    sns.set_theme(
        context=sns_context,
        style=sns_style,
        palette=sns_palette,
        font=font,
        font_scale=sns_font_scale,
    )
    lm.logger.info(
        f"set: sns.set_theme(context='{sns_context}', style='{sns_style}', palette='{sns_palette}', font='{font}', font_scale={sns_font_scale})",
    )


def plot_colortable(colors, *, ncols=4):
    cell_width = 212
    cell_height = 22
    swatch_width = 48
    margin = 12

    names = list(colors)

    n = len(names)
    nrows = math.ceil(n / ncols)

    width = cell_width * 4 + 2 * margin
    height = cell_height * nrows + 2 * margin
    dpi = 72

    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(
        margin / width,
        margin / height,
        (width - margin) / width,
        (height - margin) / height,
    )
    ax.set_xlim(0, cell_width * 4)
    ax.set_ylim(cell_height * (nrows - 0.5), -cell_height / 2.0)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()

    for i, name in enumerate(names):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(
            text_pos_x,
            y,
            name,
            fontsize=14,
            horizontalalignment="left",
            verticalalignment="center",
        )

        ax.add_patch(
            Rectangle(
                xy=(swatch_start_x, y - 9),
                width=swatch_width,
                height=18,
                facecolor=name,
                edgecolor="0.7",
            ),
        )
    plt.show()
    plt.close()


def set_figsize(figsize=(3.5, 2.5)):
    """Set the figure size for matplotlib.

    Defined in :numref:`sec_calculus`
    """
    # use_svg_display()
    plt.rcParams["figure.figsize"] = figsize


def set_axes(axes, xlabel, ylabel, xlim, ylim, xscale, yscale, legend):
    """Set the axes for matplotlib.

    Defined in :numref:`sec_calculus`
    """
    axes.set_xlabel(xlabel)
    axes.set_ylabel(ylabel)
    axes.set_xscale(xscale)
    axes.set_yscale(yscale)
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)

    if legend:
        axes.legend(legend)
    axes.grid()


def plot(
    X,
    Y=None,
    xlabel=None,
    ylabel=None,
    legend=None,
    xlim=None,
    ylim=None,
    xscale="linear",
    yscale="linear",
    fmts=("-", "m--", "g-.", "r:"),
    figsize=(3.5, 2.5),
    axes=None,
):
    """Plot data points.

    Defined in :numref:`sec_calculus`
    """
    if legend is None:
        legend = []
    def has_one_axis(X):  # True if X (tensor or list) has 1 axis
        return (hasattr(X, "ndim") and X.ndim == 1) or (
            isinstance(X, list) and not hasattr(X[0], "__len__")
        )

    if has_one_axis(X):
        X = [X]
    if Y is None:
        X, Y = [[]] * len(X), X
    elif has_one_axis(Y):
        Y = [Y]
    if len(X) != len(Y):
        X = X * len(Y)

    set_figsize(figsize)
    if axes is None:
        axes = plt.gca()
    axes.cla()
    for x, y, fmt in zip(X, Y, fmts, strict=False):
        axes.plot(x, y, fmt) if len(x) else axes.plot(y, fmt)
    set_axes(axes, xlabel, ylabel, xlim, ylim, xscale, yscale, legend)
    # return axes


def plot_dna_features(
    df: pd.DataFrame | None = None,
    use_demo_data=True,
    fig_width=10,
    fig_height=2,
    col_locus_start="locus_start",
    col_locus_length="locus_length",
    col_group="group",
    col_name="name",
    col_start="start",
    col_end="end",
    col_strand="strand",
    col_color="color",
    **kwargs,
):
    # """
    # ![](https://raw.githubusercontent.com/hermanzhaozzzz/PicturesBed01/master/PicGo/202406260156158.png)
    # ==================== DEMO ====================
    #     group     name    type  start    end  strand    color  locus_length  locus_start
    # 0    0000  AAAAAAA     CDS   1050   4000     1.0  #ffd700         20000            0
    # 1    0000     BBBB     CDS   3500   6000    -1.0  #ffcccc         20000            0
    # 2    0000       CC     CDS  14000  16000     NaN  #cffccc         20000            0
    # 3    0000     None      DR  10300  10336     NaN    black         20000            0
    # 4    0000     None      DR  10366  10402     NaN    black         20000            0
    # ..    ...      ...     ...    ...    ...     ...      ...           ...          ...
    # 510  4444     None  spacer  12106  12136     NaN      red         14000         1000
    # 511  4444     None  spacer  12172  12202     NaN      red         14000         1000
    # 512  4444     None  spacer  12238  12268     NaN      red         14000         1000
    # 513  4444     None  spacer  12304  12334     NaN      red         14000         1000
    # 514  4444     None  spacer  12370  12400     NaN      red         14000         1000

    # [515 rows x 9 columns]
    # ==================== /DEMO ====================
    # """
    def _get_demo_data():
        """Get demo data for DNA features.

        :return: demo DNA features
        :rtype: pd.DataFrame
        """
        df = pd.DataFrame(
            {
                "name": ["AAAAAAA", "BBBB", "CC"] + [None] * 50 + [None] * 50,
                "type": ["CDS", "CDS", "CDS"] + ["DR"] * 50 + ["spacer"] * 50,
                "start": (
                    [1050, 3500, 14000, *list(range(10300, 10300 + 66 * 50, 66)), *list(range(10336, 10336 + 66 * 50, 66))]
                ),
                "end": (
                    [4000, 6000, 16000, *list(range(10336, 10336 + 66 * 50, 66)), *list(range(10366, 10366 + 66 * 50, 66))]
                ),
                "strand": [1, -1, None] + [None] * 50 + [None] * 50,
                "color": (
                    [
                        "#ffd700",
                        "#ffcccc",
                        "#cffccc",
                    ]
                    + ["black"] * 50
                    + ["red"] * 50
                ),
                "locus_length": [20000] * 103,
                "locus_start": [0] * 103,
            },
        )
        ls = []
        for i in range(5):
            tmpdf = df.copy()
            tmpdf.insert(0, "group", str(i) * 4)
            if i in (2, 4):
                tmpdf["start"] = tmpdf["start"] - 300 * i
                tmpdf["end"] = tmpdf["end"] - 300 * i
                tmpdf["locus_length"] = tmpdf["locus_length"] - 1500 * i
                tmpdf["locus_start"] = 1000
            ls.append(tmpdf)
        df = pd.concat(ls)
        df = df.reset_index(drop=True)
        print("=" * 20, "DEMO", "=" * 20)
        print(df)
        print("=" * 20, "/DEMO", "=" * 20)
        return df

    if df is None:
        if use_demo_data:
            df = _get_demo_data()
        else:
            msg = "Please provide a DataFrame with DNA features."
            raise ValueError(msg)

    # 设置子图的数量
    n = df[col_group].nunique()
    # 创建一个n行1列的子图
    fig, axs = plt.subplots(n, 1, figsize=(fig_width, fig_height * n))
    if n == 1:
        # 确保 axs 是一个列表，方便统一处理
        axs = [axs]

    # 循环遍历每个子图并进行绘图
    try:
        from dna_features_viewer import GraphicFeature, GraphicRecord
    except ImportError:
        lm.logger.exception("Please install dna_features_viewer to use this function.")
        sys.exit(0)
    for i, g in zip(range(n), df.groupby(col_group), strict=False):
        group_name, data = g
        # print(f"group_name = {group_name}")

        features = []
        for _, row in data.iterrows():
            label = row.get(col_name, None)
            if isinstance(label, float):
                label = None
            start = row.get(col_start, -100)
            end = row.get(col_end, 0)
            strand = row.get(col_strand, 0)
            color = row.get(col_color, "grey")
            feature = GraphicFeature(
                start=start,
                end=end,
                strand=strand,
                color=color,
                label=label,
            )
            features.append(feature)
        if use_demo_data:
            locus_start = data["locus_start"].values[0]
            locus_length = data["locus_length"].values[0]
        else:
            locus_start = data[col_locus_start].values[0]
            locus_length = data[col_locus_length].values[0]
        record = GraphicRecord(
            first_index=locus_start,
            sequence_length=locus_length,
            features=features,
        )
        # print(axs)
        # print(n)
        record.plot(  # https://github.com/Edinburgh-Genome-Foundry/DnaFeaturesViewer
            ax=axs[i],
            figure_width=fig_width,
            draw_line=kwargs.get("draw_line", True),
            with_ruler=kwargs.get("with_ruler", True),
            plot_sequence=False,
            annotate_inline=kwargs.get("annotate_inline", True),
            max_label_length=kwargs.get("max_label_length", 50),
            max_line_length=kwargs.get("max_line_length", 20),
            level_offset=0,
            strand_in_label_threshold="default",
            elevate_outline_annotations="default",
            figure_height=fig_height,
            sequence_params=None,
        )
        # ax.set_title(f"Subplot {i+1}")  # 设置子图的标题
        axs[i].set_xlabel(
            group_name,
        )  # 设置x轴标签
        # ax.set_ylabel("Y Label")  # 设置y轴标签
        axs[i].grid(False)  # 显示网格线
        # plt.show()
    plt.tight_layout()
    return fig

    # 调整子图之间的距离


if __name__ == "__main__":
    pass
    # test for init_matplotlib
    # init_matplotlib(log_level="warning")
    # plot_colortable(
    #     ["#64C1E8", "#80CED7", "#63C7B2", "#8E6C88", "#CA61C3", "#FF958C", "#883677"],
    #     ncols=1,
    # )

    # def normal(x, mu, sigma):
    #     p = 1 / math.sqrt(2 * math.pi * sigma**2)
    #     return p * np.exp(-0.5 / sigma**2 * (x - mu) ** 2)

    # x = np.arange(-7, 7, 0.01)
    # params = [(0, 1), (0, 2), (3, 1)]
    # plot(
    #     x,
    #     [normal(x, mu, sigma) for mu, sigma in params],
    #     xlabel="x",
    #     ylabel="p(x)",
    #     figsize=(4.5, 2.5),
    #     legend=[f"mean {mu}, std {sigma}" for mu, sigma in params],
    # )
    # plt.show()
    # plt.close()
    # df_plot = pd.read_csv("~/demo.tsv", sep="\t")
    # # print(df_plot)
    # plot_dna_features()
    # plot_dna_features(
    #     df=df_plot,
    #     # use_demo_data=True
    #     fig_width=20,
    #     fig_height=3,
    #     col_locus_start="locus_start",
    #     col_locus_length="locus_length",
    #     col_group="crispr_id:member",
    #     col_name="pep_name",
    #     col_start="pep_start",
    #     col_end="pep_end",
    #     col_strand="pep_strand",
    #     col_color="color",
    # )

    # # fig.savefig("/Users/zhaohuanan/Downloads/test.png", dpi=300)
    # # 显示图形
    # # plt.show()



"""_summary_.

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

_description_

example 1:
    bioat list
        <in shell>:
            $ bioat list
        <in python consolo>:
            >>> from bioat.cli import Cli
            >>> bioat = Cli()
            >>> bioat.list()
            >>> print(bioat.list())

example 2:
    _example_
"""

import sys

import pandas as pd

from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.tabletools")


class TableTools:
    """To integrate tables."""

    lm.set_names(cls_name="TableTools")

    def __init__(self):
        pass

    def merge(
        self,
        inputs,
        tags,
        output,
        input_fmt="tsv",
        output_fmt="tsv",
        input_header=False,
        output_header=False,
        log_level="WARNING",
    ):
        """A simple tool to merge same formatted tables from different sample.

        Params
        :param inputs: table files
        :param tags: tags for each table
        :param output: merged file
        :param input_fmt: tsv | csv
        :param output_fmt: tsv | csv
        :param input_header: True | False, input table has header or not
        :param output_header: True | False, output table has header or not
        :param log_level: log status
        """
        lm.set_names(func_name="merge")
        lm.set_level(log_level)

        # fix params
        inputs = list(inputs) if isinstance(inputs, tuple) else inputs.split(",")
        tags = list(tags) if isinstance(tags, tuple) else tags.split(",")

        lm.logger.debug(f"input_fmt={input_fmt}")
        lm.logger.debug(f"output_fmt={output_fmt}")
        lm.logger.debug(f"input_header={input_header}")
        lm.logger.debug(f"output_header={output_header}")

        input_header = 0 if input_header else None
        output = sys.stdout if output == sys.stdout.name else output
        dt_sep = {"csv": ",", "tsv": "\t"}

        df = pd.DataFrame()

        for file, tag in zip(inputs, tags, strict=False):
            _df = pd.read_csv(
                file, sep=dt_sep[input_fmt], header=input_header, index_col=None
            )
            _df.insert(0, column="<sample>", value=tag)
            df = pd.concat([df, _df], axis=0, ignore_index=True)

        df.to_csv(output, sep=dt_sep[output_fmt], header=output_header, index=None)

    def split(
        self,
        input: str,
        n: int,
        output_prefix=None,
        input_fmt="tsv",
        output_fmt="tsv",
        input_header=False,
        output_header=False,
        compress=False,
        log_level="WARNING",
    ):
        """A simple tool to split table into parts.

        Params
        :param input: table to split
        :param n: split table into n parts
        :param output_prefix: name prefix for splitted parts, the same with input if not defined
        :param input_fmt: tsv | csv
        :param output_fmt: tsv | csv
        :param input_header: True | False, input table has header or not
        :param output_header: True | False, output table has header or not
        :param compress: True | False, gzip the output table or not
        :param log_level: log status
        """
        lm.set_names(func_name="split")
        lm.set_level(log_level)

        lm.logger.debug(f"input_fmt={input_fmt}")
        lm.logger.debug(f"output_fmt={output_fmt}")
        lm.logger.debug(f"input_header={input_header}")
        lm.logger.debug(f"output_header={output_header}")
        # fix params
        input_header = 0 if input_header else None
        output_prefix = output_prefix if output_prefix else f"{input}_"
        compress = ".gz" if compress else ""

        dt_sep = {"csv": ",", "tsv": "\t"}
        # load table
        df = pd.read_csv(input, header=input_header, sep=dt_sep[input_fmt])

        # max lines for each part table: chunk_size
        chunk_size = df.shape[0] // n if df.shape[0] % n == 0 else df.shape[0] // n + 1

        # write out
        for idx in range(n):
            df_write_out = df.iloc[idx * chunk_size : (idx + 1) * chunk_size,]
            df_write_out.to_csv(
                f"{output_prefix}{idx}.%s{compress}" % output_fmt,
                index=None,
                header=output_header,
                sep=dt_sep[output_fmt],
            )


if __name__ == "__main__":
    pass



import os


def check_cmd(x):
    return any(
        os.access(os.path.join(path, x), os.X_OK)
        for path in os.environ["PATH"].split(os.pathsep)
    )


def print_head(SAMPLES, MODE):
    print("----------\nSAMPLES:")
    [print("\t" + i) for i in SAMPLES]
    print("----------\nMODE:")
    print("\t" + MODE)
    print("----------\n\n")


def check_read(x):
    if x == "PE":
        read = ["R1", "R2"]
    elif x == "SE":
        read = ["SE"]
    else:
        raise ValueError()
    return read
